#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <pthread.h>
#include <linux/fb.h>
#include <linux/videodev2.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/poll.h>

#include <gst/gst.h>
#include <gst/app/gstappsink.h>

#define VIN_MAX		1

typedef struct _app_data {
	int pipefd[2];
	pthread_t tid;
	GMainLoop *loop;
	GstElement *pipeline;
} app_data;


static void on_bus_message(GstBus *bus, GstMessage *message, gpointer appdata)
{
	app_data *d = (app_data *)appdata;

	switch (GST_MESSAGE_TYPE(message)) {
	case GST_MESSAGE_ERROR:
		g_print("ERROR received\n");
		g_main_loop_quit(d->loop);
		break;
	case GST_MESSAGE_EOS:
		g_print("EOS reached\n");
		g_main_loop_quit(d->loop);
		break;
	default:
		break;
	}
}

void *event_thread_func(void *appdata)
{
	app_data *d = (app_data *)appdata;
	struct pollfd fds[VIN_MAX + 1];
	char vdev[15];
	unsigned int vin, i;
	int vfd;

	struct v4l2_event_subscription sub = {0};
	sub.type = V4L2_EVENT_SOURCE_CHANGE;

	for (vin = 0; vin < VIN_MAX; vin++) {
		snprintf(vdev, sizeof(vdev), "/dev/video%u", vin);
		vfd = open(vdev, O_RDONLY | O_NONBLOCK);
		if (vfd < 0) {
			g_print("%s open failed: %s\n", vdev, g_strerror(errno));
			goto out;
		}

		if (ioctl(vfd, VIDIOC_SUBSCRIBE_EVENT, &sub) < 0) {
			g_print("VIDIOC_SUBSCRIBE_EVENT failed: %s\n", g_strerror(errno));
			goto out;
		}

		fds[vin].fd = vfd;
		fds[vin].events = POLLPRI;
	}

	fds[VIN_MAX].fd = d->pipefd[0];
	fds[VIN_MAX].events = POLLIN;

	while (1) {
		if (poll(fds, VIN_MAX + 1, -1) < 0) {
			g_print("poll failed: %s\n", g_strerror(errno));
			break;
		}

		if (fds[VIN_MAX].revents & POLLIN) {
			g_print("QUIT message received\n");
			break;
		}

		for (i = 0; i < VIN_MAX; i++) {
			if (fds[i].revents & POLLPRI) {
				struct v4l2_event ev = {0};

				if (ioctl(fds[i].fd, VIDIOC_DQEVENT, &ev) == 0 &&
						ev.type == V4L2_EVENT_SOURCE_CHANGE) {
					g_print("Source Change event vin = %u\n", i);

					g_print("Pipeline stoppped\n");
					gst_element_set_state(d->pipeline, GST_STATE_NULL);
					if (gst_element_get_state (d->pipeline, NULL, NULL, -1) == GST_STATE_CHANGE_FAILURE) {
						g_error ("Failed to go into NULL state");
					}

					g_print("Pipeline running\n");
					gst_element_set_state(d->pipeline, GST_STATE_PLAYING);
					if (gst_element_get_state (d->pipeline, NULL, NULL, -1) == GST_STATE_CHANGE_FAILURE) {
						g_error ("Failed to go into PLAYING state");
					}
				}
			}
		}
	}

out:
	for (i = 0; i < vin; i++)
		close(fds[i].fd);
}

int event_thread_create(app_data *d)
{
	if (pipe(d->pipefd) < 0) {
		g_print("pipe failed: %s\n", g_strerror(errno));
		return (-1);
	}
	return pthread_create(&d->tid, NULL, event_thread_func, (void *)d);
}

int event_thread_cancel(app_data *d)
{
	if (write(d->pipefd[1], "q", 1) != 1) {
		g_print("could not send message to event thread\n");
		return (-1);
	}

	if (pthread_join(d->tid, NULL) != 0) {
		g_print("failed to join event thread\n");
		return (-1);
	}

	close(d->pipefd[0]);
	close(d->pipefd[1]);
	return 0;
}

int main(int argc, char *argv[])
{
	app_data *data = g_new0(app_data, 1);
	GError *error = NULL;
	GstElement *sink;
	GstBus *bus;

	gst_init(&argc, &argv);
	data->pipeline = gst_parse_launch(
		"imxv4l2videosrc device=/dev/video0 ! imxg2dvideosink", &error);
	if (error) {
		g_print("Creating pipeline failed: %s\n", error->message);
		g_free(data);
		return (-1);
	}

	bus = gst_pipeline_get_bus(GST_PIPELINE(data->pipeline));
	gst_bus_add_signal_watch(bus);
	g_signal_connect(bus, "message", (GCallback)on_bus_message, (gpointer)data);
	gst_object_unref(bus);

	g_print("Pipeline running\n");
	gst_element_set_state(data->pipeline, GST_STATE_PLAYING);
	data->loop = g_main_loop_new(NULL, FALSE);

	if (event_thread_create(data) == 0) {
		g_main_loop_run(data->loop);
		event_thread_cancel(data);
	}

	g_print("Pipeline stoppped\n");
	gst_element_set_state(data->pipeline, GST_STATE_NULL);

	gst_object_unref(data->pipeline);
	g_main_loop_unref(data->loop);
	g_free(data);

	return 0;
}
